/***************************** C SOURCE FILE *********************************
**
**  Project:    Logic IO RTCU Gateway Protocol
**  Filename:   GWPROTLD.C 
**  
******************************************************************************
**
**  This is the the TCP/IP layer for gwprot protocol running on a Windows PC.
**
*****************************************************************************/


/*****************************************************************************
**
** MODULES USED 
**
*****************************************************************************/
#define WIN32_LEAN_AND_MEAN      // Exclude rarely-used stuff from Windows headers
#include <winsock2.h>
#include <windows.h>
#include "gwprotld.h"
#include <conio.h>
#include <stdlib.h>
#include <openssl/bio.h>
#include <openssl/ssl.h>


/*****************************************************************************
**
**    DEFINITIONS AND MACROS
**
*****************************************************************************/
#define ENTER_CRIT          EnterCriticalSection(&conn->lock)
#define EXIT_CRIT           LeaveCriticalSection(&conn->lock)


/****************************************************************************
**
**    TYPEDEFS AND STRUCTURES
**
****************************************************************************/
typedef struct sock_inst_t {
   CRITICAL_SECTION       lock;
   SOCKET                 sock;
   SSL_CTX*               tls_ctx;
   SSL*                   tls_ssl;
   long                   cookie;
} sock_inst_t;


/*****************************************************************************
**
** EXPORTED VARIABLES 
**
*****************************************************************************/


/*****************************************************************************
**
** GLOBAL VARIABLES 
**
*****************************************************************************/
void gwInitialize(void);


/*****************************************************************************
**
** LOCAL FUNCTIONS 
**
*****************************************************************************/

/////////////////////////////////////////////////////////////////////////////
// This function will clean up TLS connection and context
//
// Input:
//      conn             - The session to clean up
//
// Output:
//      None.
//
// Returns:
//      None.
//
/////////////////////////////////////////////////////////////////////////////
static void CleanUpTls(sock_inst_t* conn)
{
   // Clean up session
   if (conn->tls_ssl)
   {
      SSL_free(conn->tls_ssl);
      conn->tls_ssl = NULL;
   }
   // Clean up context
   if (conn->tls_ctx)
   {
      SSL_CTX_free(conn->tls_ctx);
      conn->tls_ctx = NULL;
   }
}

/////////////////////////////////////////////////////////////////////////////
// This function will attempt to connect to the gateway using the specified
// IP-address and port. The IP-address can be symbolic, like "gw.rtcu.dk"
//
// Input:
//      ip               - Zero terminate ASCII string with the IP-address.
//      port             - The port where the gateway listens for clients.
//      conn             - Socket if !INVALID_SOCKET then it will be tried
//                         closed, before new connection is created.
//      server           - Zero terminated ASCII string with the path to
//                         the PEM file needed to verify the server certificate.
//      client_cert      - Zero terminated ASCII string with the path to
//                         the PEM file with the client certificate.
//      client_key       - Zero terminated ASCII string with the path to
//                         the PEM file private key of the client certificate.
//      client_pass      - Zero terminated ASCII string with the password of
//                         the private key.
//
// Output:
//      None.
//
// Returns:
//      0                - Success.
//      1                - Error connecting.
//
/////////////////////////////////////////////////////////////////////////////
static unsigned short doConnect(
                                sock_inst_t*   conn,
                                const char*    ip,
                                unsigned short port,
                                const char*    server,
                                const char*    client_cert,
                                const char*    client_key,
                                const char*    client_pass
                               )
{
   // Var
   unsigned long      addr;
   struct sockaddr_in sock_addr;
   struct hostent*    hp;

   // release socket
   gwldDisconnect(conn);

   // Create a socket
   conn->sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
   if (conn->sock == INVALID_SOCKET) return 1;

   // Try to resolve address, either as a "real" ip address, or as a symbolic name
   addr = inet_addr(ip);
   if (addr == INADDR_NONE) {
      hp = gethostbyname(ip);
      if (hp == NULL) {
         closesocket(conn->sock);
         conn->sock = INVALID_SOCKET;
         return 1;
      }
      addr = *((unsigned long*)hp->h_addr);
   }

   // Now connect to the server (using ip and port)
   sock_addr.sin_addr.s_addr = addr;
   sock_addr.sin_family      = AF_INET;
   sock_addr.sin_port        = htons(port);
   if (connect(conn->sock, (struct sockaddr*) & sock_addr, sizeof(sock_addr))) {
      closesocket(conn->sock);
      conn->sock = INVALID_SOCKET;
      return 1;
   }

   // Enable security
   if (server && strlen(server))
   {
      // Var
      WSAPOLLFD fdset;
      X509*     cert;
      int       rc, iMode = 1, timeout = 30;

      // Change mode to nonblocking
      if (ioctlsocket(conn->sock, FIONBIO, (u_long FAR*) & iMode) == SOCKET_ERROR)
      {
         // Clean up
         closesocket(conn->sock);
         conn->sock = INVALID_SOCKET;

         // Completed
         return 2;
      }

      // Create TLS context
      conn->tls_ctx = SSL_CTX_new(TLS_method());
      if (conn->tls_ctx == NULL)
      {
         closesocket(conn->sock);
         conn->sock = INVALID_SOCKET;
         return 2;
      }

      // Configure TLS context
      rc = SSL_CTX_set_min_proto_version(conn->tls_ctx, TLS1_1_VERSION);
      if (rc != 1)
      {
         // Clean up
         CleanUpTls(conn);
         closesocket(conn->sock);
         conn->sock = INVALID_SOCKET;

         // Completed
         return 2;
      }
      rc = SSL_CTX_load_verify_locations(conn->tls_ctx, server, NULL);
      if (rc != 1)
      {
         // Clean up
         CleanUpTls(conn);
         closesocket(conn->sock);
         conn->sock = INVALID_SOCKET;

         // Completed
         return 2;
      }
      rc = SSL_CTX_set_cipher_list(conn->tls_ctx, "ALL:!ADH:!LOW:!EXP:!MD5:@STRENGTH");
      if (rc != 1)
      {
         // Clean up
         CleanUpTls(conn);
         closesocket(conn->sock);
         conn->sock = INVALID_SOCKET;

         // Completed
         return 2;
      }
	  if (client_cert && strlen(client_cert)) {
         // Set password
         SSL_CTX_set_default_passwd_cb_userdata(conn->tls_ctx, client_pass);

         // Set certificate
         if (!SSL_CTX_use_certificate_file(conn->tls_ctx, client_cert, SSL_FILETYPE_PEM)
             || !SSL_CTX_use_PrivateKey_file(conn->tls_ctx, client_key, SSL_FILETYPE_PEM)
             || !SSL_CTX_check_private_key(conn->tls_ctx))
         {
            // Clean up
            CleanUpTls(conn);
            closesocket(conn->sock);
            conn->sock = INVALID_SOCKET;

            // Completed
            return 2;
         }
      }

      // Create SSL
      conn->tls_ssl = SSL_new(conn->tls_ctx);
      if (conn->tls_ssl == NULL) {
         // Clean up
         CleanUpTls(conn);
         closesocket(conn->sock);
         conn->sock = INVALID_SOCKET;

         // Completed
         return 3;
      }
      rc = SSL_set_fd(conn->tls_ssl, conn->sock);
      if (rc != 1)
      {
         // Clean up
         CleanUpTls(conn);
         closesocket(conn->sock);
         conn->sock = INVALID_SOCKET;

         // Completed
         return 3;
      }

      // TLS handshake
      rc = SSL_connect(conn->tls_ssl);
      while (rc < 1)
      {
         // Handle error
         rc = SSL_get_error(conn->tls_ssl, rc);
         if (rc != SSL_ERROR_WANT_READ && rc != SSL_ERROR_WANT_WRITE)
         {
            // Clean up
            CleanUpTls(conn);
            closesocket(conn->sock);
            conn->sock = INVALID_SOCKET;

            // Completed
            return 4;
         }

         // Setup
         fdset.fd = conn->sock;
         fdset.events = (rc == SSL_ERROR_WANT_READ ? POLLIN : 0) | (rc == SSL_ERROR_WANT_WRITE ? POLLOUT : 0);
         fdset.revents = 0;

          // Check state
         rc = WSAPoll(&fdset, 1, 1000);
         if (rc < 0 || (rc && fdset.revents & (POLLNVAL | POLLHUP | POLLERR)))
         {
            // Clean up
            CleanUpTls(conn);
            closesocket(conn->sock);
            conn->sock = INVALID_SOCKET;

            // Completed
            return 4;
         }
         if (!rc && timeout-- < 1)
         {
            // Clean up
            CleanUpTls(conn);
            closesocket(conn->sock);
            conn->sock = INVALID_SOCKET;

            // Completed
            return 4;
         }

         // Try again
         rc = SSL_connect(conn->tls_ssl);
      }

      // Verify server certificate
      cert = SSL_get_peer_certificate(conn->tls_ssl);
      if (cert) { X509_free(cert); }
      else
      {
         // Clean up
         CleanUpTls(conn);
         closesocket(conn->sock);
         conn->sock = INVALID_SOCKET;

         // Completed
         return 5;
      }
      rc = SSL_get_verify_result(conn->tls_ssl);
      if (rc != X509_V_OK)
      {
         // Clean up
         CleanUpTls(conn);
         closesocket(conn->sock);
         conn->sock = INVALID_SOCKET;

         // Completed
         return 6;
      }
   }

   // Completed
   return 0;
}


/*****************************************************************************
**
** EXPORTED FUNCTIONS 
**
*****************************************************************************/

/////////////////////////////////////////////////////////////////////////////
// Initialize DLL
//
// Input:
//      hModule
//      ul_reason_for_call
//      lpReserved
//
// Output:
//      None.
//
// Returns:
//      None.
//
/////////////////////////////////////////////////////////////////////////////
BOOL APIENTRY DllMain(
                      HANDLE hModule, 
                      DWORD  ul_reason_for_call, 
                      LPVOID lpReserved
                     )
{
   // Var
   WSADATA wsaData;

   // Handle reason
   switch (ul_reason_for_call) {
      case DLL_PROCESS_ATTACH:
         // Initialize module
         gwInitialize();

         // Initialize WinSocket
         WSAStartup(0x202, &wsaData);
         break;

      case DLL_PROCESS_DETACH:
         // Release WinSocket
         WSACleanup();
         break;

      case DLL_THREAD_ATTACH:
      case DLL_THREAD_DETACH:
         break;
   }

   // Completed
   return TRUE;
}

/////////////////////////////////////////////////////////////////////////////
// Create a new socket, this function must be called before any other gwprotld
// functions.
//
// Input:
//      None.
//
// Output:
//      conn             - Pointer to socket instance.
//
// Returns:
//      0                - Success.
//      1                - Error connecting.
//
/////////////////////////////////////////////////////////////////////////////
unsigned short gwldCreate(sock_inst_t ** conn)
{
   // Validate
   if (!conn) return 1;
   if (*conn && (*conn)->cookie == 0x343499BA) return 1;

   // Allocate instance structure
   *conn = (sock_inst_t *)malloc(sizeof(sock_inst_t));
   if (*conn == NULL) return 1;

   // Initialize instance
   InitializeCriticalSection(&(*conn)->lock);
   (*conn)->sock     = INVALID_SOCKET;
   (*conn)->cookie   = 0x343499BA;
   (*conn)->tls_ssl  = NULL;
   (*conn)->tls_ctx  = NULL;

   // Successful
   return 0;
}

/////////////////////////////////////////////////////////////////////////////
// Delete a socket, after this function is called no other gwprotld function
// will work.
//
// Input:
//      None.
//
// Output:
//      conn             - Pointer to socket instance.
//
// Returns:
//      0                - Success.
//      1                - Error connecting.
//
/////////////////////////////////////////////////////////////////////////////
unsigned short gwldDestroy(sock_inst_t ** conn)
{
   // Validate
   if (!*conn) return 1;
   if ((*conn)->cookie != 0x343499BA) return 1;

   // Cleanup
   gwldDisconnect((*conn));
   DeleteCriticalSection(&(*conn)->lock);
   (*conn)->sock   = INVALID_SOCKET;
   (*conn)->cookie = 0;

   // De-allocate session structure
   free(*conn);
   *conn = NULL;

   // Completed
   return 0;
}

/////////////////////////////////////////////////////////////////////////////
// This function will attempt to connect to the gateway using the specified
// IP-address and port. The IP-address can be symbolic, like "gw.rtcu.dk"
//
// Input:
//      conn             - Pointer to socket instance.
//      ip               - Zero terminate ASCII string with the IP-address.
//      port             - The port where the gateway listens for clients.
//      conn             - Socket if !INVALID_SOCKET then it will be tried
//                         closed, before new connection is created.
//      server           - Zero terminated ASCII string with the path to
//                         the PEM file needed to verify the server certificate.
//      client_cert      - Zero terminated ASCII string with the path to
//                         the PEM file with the client certificate.
//      client_key       - Zero terminated ASCII string with the path to
//                         the PEM file private key of the client certificate.
//      client_pass      - Zero terminated ASCII string with the password of
//                         the private key.
//
// Output:
//      None.
//
// Returns:
//      0                - Success.
//      1                - Error connecting.
//
/////////////////////////////////////////////////////////////////////////////
unsigned short gwldConnect(
                           sock_inst_t*   conn,
                           const char*    ip,
                           unsigned short port,
                           const char*    server,
                           const char*    client_cert,
                           const char*    client_key,
                           const char*    client_pass
                          )
{
   // Var
   unsigned short rc = 0;

   // Validate
   if (conn == NULL) return 1;
   if (conn->cookie != 0x343499BA) return 1;

   // Lock
   ENTER_CRIT;

   // create connection
   rc = doConnect(conn, ip, port, server, client_cert, client_key, client_pass);

   // Unlock
   EXIT_CRIT;

   // Successful
   return rc;
}

/////////////////////////////////////////////////////////////////////////////
// This function will disconnect from the gateway.
//
// Input:
//      conn             - Pointer to socket instance.
//
// Output:
//      None.
//
// Returns:
//      0                - Success.
//      1                - Error disconnecting.
//
/////////////////////////////////////////////////////////////////////////////
unsigned short gwldDisconnect(sock_inst_t *conn)
{
   // Verify
   if (conn == NULL) return 1;
   if (conn->cookie != 0x343499BA) return 1;

   // Lock
   ENTER_CRIT;

   // Close socket
   if (conn->tls_ssl)
   {
      // Var
      WSAPOLLFD fdset;
      int       rc, timeout = 10;

      // Close session
      rc = SSL_shutdown(conn->tls_ssl);
      while (rc < 1)
      {
         // Handle error
         rc = SSL_get_error(conn->tls_ssl, rc);
         if (rc != SSL_ERROR_WANT_READ && rc != SSL_ERROR_WANT_WRITE) break;

         // Setup
         fdset.fd = conn->sock;
         fdset.events = (rc == SSL_ERROR_WANT_READ ? POLLIN : 0) | (rc == SSL_ERROR_WANT_WRITE ? POLLOUT : 0);
         fdset.revents = 0;

         // Check state
         rc = WSAPoll(&fdset, 1, 1000);
         if (rc < 0 || (rc && fdset.revents & (POLLNVAL | POLLHUP | POLLERR))) break;
         if (!rc && timeout-- < 1) break;

         // Try again
         rc = SSL_shutdown(conn->tls_ssl);
      }
   }
   if (conn->sock != INVALID_SOCKET)
   {
      closesocket(conn->sock);
      conn->sock = INVALID_SOCKET;
   }

   // Clean up
   CleanUpTls(conn);

   // Unlock
   EXIT_CRIT;

   // Completed
   return 0;
}

/////////////////////////////////////////////////////////////////////////////
// This function will disconnect the connection to the internet.
// It is assumed that the line driver will automatically reconnect.
//
// Input:
//      conn             - Pointer to socket instance.
//
// Output:
//      None.
//
// Returns:
//      0                - Success.
//      1                - Error hanging up internet connection.
//
/////////////////////////////////////////////////////////////////////////////
unsigned short gwldDisconnectInternet(sock_inst_t *conn)
{
   // Dummy
   conn = conn;
   return 0;
}

/////////////////////////////////////////////////////////////////////////////
// This function will query the line driver whether the connection to the
// internet is present.
//
// Input:
//      conn             - Pointer to socket instance.
//
// Output:
//      None.
//
// Returns:
//      0                - Internet is available.
//      1                - Internet is not available.
//
/////////////////////////////////////////////////////////////////////////////
unsigned short gwldIsInternetPresent(sock_inst_t *conn)
{
   // Dummy
   conn = conn;
   return 0;
}

/////////////////////////////////////////////////////////////////////////////
// This function will query the line driver whether the connection to the
// internet is temporarily suspended.
//
// Input:
//      conn             - Pointer to socket instance.
//
// Output:
//      None.
//
// Returns:
//      0                - Internet is suspended.
//      1                - Internet is not suspended.
//
/////////////////////////////////////////////////////////////////////////////
unsigned short gwldIsInternetSuspended(sock_inst_t *conn)
{
   // Dummy
   conn = conn;
   return 1;
}

/////////////////////////////////////////////////////////////////////////////
// This function will send data to the gateway.
//
// Input:
//      conn             - Pointer to socket instance.
//      data             - Pointer to buffer containing the data to send.
//      data_len         - The number of bytes to send.
//
// Output:
//      None.
//
// Returns:
//      0                - Data was sent successfully.
//      1                - Data was not sent successfully.
//
/////////////////////////////////////////////////////////////////////////////
unsigned short gwldSend(
                        sock_inst_t*   conn,
                        const void*    data,
                        unsigned short data_len
                       )
{
   // Var
   int rc;

   // Validate
   if (conn == NULL) return 1;
   if (conn->cookie != 0x343499BA) return 1;

   // Write data
   if (conn->tls_ssl)
   {
      // Var
      WSAPOLLFD fdset;

      // Lock
      ENTER_CRIT;

      if (conn->tls_ssl == NULL)
      {
         // Unlock
         EXIT_CRIT;

         // Completed
         return 1;
      }

      // Write
      rc = SSL_write(conn->tls_ssl, data, data_len);
      while (rc < 1)
      {
         // Handle error
         rc = SSL_get_error(conn->tls_ssl, rc);
         EXIT_CRIT;
         if (rc != SSL_ERROR_WANT_READ && rc != SSL_ERROR_WANT_WRITE)
         {
            return 1;
         }

         // Setup
         fdset.fd = conn->sock;
         fdset.events = (rc == SSL_ERROR_WANT_READ ? POLLIN : 0) | (rc == SSL_ERROR_WANT_WRITE ? POLLOUT : 0) | POLLNVAL;
         fdset.revents = 0;

         // Check state
         rc = WSAPoll(&fdset, 1, 1000);
         if (rc < 0 || (rc > 0 && !(fdset.revents & (POLLIN | POLLOUT))))
         {
            return 1;
         }

         // Lock
         ENTER_CRIT;

         if (conn->tls_ssl == NULL)
         {
            // Unlock
            EXIT_CRIT;

            // Completed
            return 1;
         }

         // Try again
         rc = SSL_write(conn->tls_ssl, data, data_len);
      }

      // Unlock
      EXIT_CRIT;
   }
   else
   {
      rc = send(conn->sock, (char*)data, data_len, 0);
   }

   // Completed
   return (rc != data_len) ? 1 : 0;
}

/////////////////////////////////////////////////////////////////////////////
// This function will receive data from the gateway.
// This function will block until data is available.
//
// Input:
//      conn             - Pointer to socket instance.
//      max_data_len     - The size of the data buffer.
//
// Output:
//      data             - Pointer to buffer where the received data is stored.
//      data_len         - The number of bytes received.
//
// Returns:
//      0                - Data was received successfully.
//      1                - Data was not received successfully.
//
/////////////////////////////////////////////////////////////////////////////
unsigned short gwldReceive(
                           sock_inst_t*    conn,
                           unsigned short  max_data_len,
                           void*           data,
                           unsigned short* data_len
                          )
{
   // Var
   int rc;

   // Validate
   if (conn == NULL) return 1;
   if (conn->cookie != 0x343499BA) return 1;

   // Read incoming data
   if (conn->tls_ssl)
   {
      // Var
      WSAPOLLFD fdset;
      size_t    len;

      // Lock
      ENTER_CRIT;

      if (conn->tls_ssl == NULL)
      {
         // Unlock
         EXIT_CRIT;

         // Set output
         *data_len = 0;

         // Completed
         return 1;
      }

      // Read data
      rc = SSL_read_ex(conn->tls_ssl, data, max_data_len, &len);
      while (!rc)
      {
         // Get error
         rc = SSL_get_error(conn->tls_ssl, rc);
         EXIT_CRIT;
         if (rc != SSL_ERROR_WANT_READ && rc != SSL_ERROR_WANT_WRITE)
         {
            *data_len = 0;
            return 1;
         }

         // Setup
         fdset.fd      = conn->sock;
         fdset.events  = (rc == SSL_ERROR_WANT_READ ? POLLIN : 0) | (rc == SSL_ERROR_WANT_WRITE ? POLLOUT : 0);
         fdset.revents = 0;

         // Check state
         rc = WSAPoll(&fdset, 1, 1000);
         if (rc < 0 || (rc == 1 && fdset.revents & (POLLNVAL | POLLHUP | POLLERR)))
         {
            *data_len = 0;
            return 1;
         }

         // Lock
         ENTER_CRIT;

         if (conn->tls_ssl == NULL)
         {
            // Unlock
            EXIT_CRIT;

            // Set output
            *data_len = 0;

            // Completed
            return 1;
         }

         // Try again
         rc = SSL_read_ex(conn->tls_ssl, data, max_data_len, &len);
      }

      // Unlock
      EXIT_CRIT;

      // Set output
      *data_len = len;
   }
   else
   {
      rc = recv(conn->sock, (char*)data, max_data_len, 0);
      if (rc == 0 || rc == SOCKET_ERROR) {
         *data_len = 0;
         return 1;
      }
      *data_len = rc;
   }

   // Completed
   return 0;
}


/****************************************************************************
**
**    EOF 
**
****************************************************************************/
